@@ -22,7 +22,7 @@
 
   However Display Ghostscript System depends on GLib. */
 
-/* $Id: gsdpsx.c,v 1.50 1999/09/17 19:06:56 masata-y Exp $ */
+/* $Id: gsdpsx.c,v 1.64 2000/01/10 13:27:05 masata-y Exp $ */
 /* X Windows Display PostScript additions for Ghostscript library */
 
 
@@ -66,6 +66,8 @@ typedef GSList GSet;
 
 #include "gdevx.h"
 extern gx_device_X gs_x11_device;
+#define gx_device_is_x11(dev)\
+ ((dev)->dname == gs_x11_device.dname)
 
 #include "gsdpsx.h"
 
@@ -93,6 +95,18 @@ enum conn_type { UNIXCONN = XDPSNX_TRANS_UNIX, TCPCONN =  XDPSNX_TRANS_TCP};
 #include <sys/stat.h>
 #include <sys/un.h>
 
+#if defined(HAVE_TCPD_H)
+#if defined(HAVE_SYSLOG_H)
+#include <syslog.h>
+#else
+#define LOG_INFO 6
+#define LOG_WARNING 4
+#endif
+#include "tcpd.h"
+int allow_severity = LOG_INFO;
+int deny_severity = LOG_WARNING;
+#endif
+
 #ifndef CSDPS_UNIX_PATH
 #ifdef hpux
 #define CSDPS_UNIX_DIR      "/usr/spool/sockets/DPSNX"
@@ -355,9 +369,11 @@ private const char * util_x_cap_request_enum2str(int key);
 private const char * util_signal_int2str(int sig);
 private const char * util_status_enum2str(int key);
 private const char * util_conn_type_enum2str(enum conn_type type);
-private void util_print_pid(int sec);
 private int util_byteorder_check(gsdpsx_conn_info *conn);
 
+/* debug functions */
+private void debug_print_pid(int sec);
+
 /* ------ DPS/X11 streams ------ */
 private void gsdpsx_std_init(void);
 
@@ -380,7 +396,7 @@ gs_startXdpsagent(gs_state *pgs, gs_memory_t *mem)
       DGS_DEBUG = 0;
     else if (DGS_DEBUG < 0)
       dprintf("Enter debug mode\n");
-    
+
     if (args_get_match(dgs_args, "help"))
       {
 	dprintf("Usage: dpsnx.agent [--help] [--debug=N] [--version] [tcp/PORT|unix/PORT]\n");
@@ -494,9 +510,10 @@ gs_startXdpsagent(gs_state *pgs, gs_memory_t *mem)
 
     while (1)
       {
-	int in_sock;
-	enum conn_type type;
-
+	int in_sock = -1;
+	enum conn_type type = 0;
+	int xfd = XConnectionNumber(shared_dpy);
+	
 	result = listen(server_socket_unix, 5);
 	if (result < 0)
 	  {
@@ -512,10 +529,12 @@ gs_startXdpsagent(gs_state *pgs, gs_memory_t *mem)
 	  }
 	max_port = (server_socket_unix<server_socket_tcp)?
 	  server_socket_tcp:server_socket_unix;
+	max_port = (xfd<max_port)?max_port:xfd;
     
 	FD_ZERO(&rfds);
 	FD_SET(server_socket_unix, &rfds);
 	FD_SET(server_socket_tcp, &rfds);
+	FD_SET(xfd, &rfds);
 	proc_server_setup_signal();
 	result = select(max_port+1, &rfds, NULL, NULL, NULL);
 	if (-1 == result)
@@ -542,9 +561,26 @@ gs_startXdpsagent(gs_state *pgs, gs_memory_t *mem)
 	    in_sock = conn_insock_accept(server_socket_tcp, 
 					 TCPCONN);
 	    type = TCPCONN;
+
+#if defined(HAVE_TCPD_H)        
+	    {
+	      struct request_info request;
+	      request_init(&request, RQ_DAEMON, "dgs", RQ_FILE, in_sock, 0);
+	      fromhost(&request);
+	      if (!hosts_access(&request)) {
+		dprintf3("Connection refused on socket number %d via %s from %s\n",
+			 in_sock, "TCPCONN", eval_client(&request));
+		close(in_sock);
+		continue;
+	      }
+	    }
+#endif        
+
 	    dprintf2("Connection accepted on socket number %d via %s\n", 
 		     in_sock, "TCPCONN");
 	  }
+	else if (FD_ISSET(xfd, &rfds))
+	  goto finalize_sock_tcp;
 
 	if (in_sock == -1) 
 	  goto atom_cleanup;
@@ -610,7 +646,7 @@ proc_agent_start(int in_sock,
       const char * run_string = "(%stdin) run  ";
       int len;				     /* message length */
 
-      util_print_pid(DGS_DEBUG);
+      debug_print_pid(DGS_DEBUG);
 
       /** Shutdown some server resources */
       billboard_atom_needs_cleanup = 0;	     /* atom_cleanup:  */
@@ -692,6 +728,7 @@ proc_agent_start(int in_sock,
   else 
     {
       /** Parent */
+      conn_insock_close(in_sock); 
       server_agent_pids = g_set_add(server_agent_pids, (gpointer)pid);
       return 0;
     }
@@ -780,7 +817,7 @@ proc_agent_finalize(int sig)
  
       }
       XDestroyWindow(shared_dpy, conn_info.agent_win);
-      XCloseDisplay(shared_dpy);
+      XCloseDisplay(shared_dpy); 
     }
   gs_exit(0);
 }
@@ -876,15 +913,16 @@ private Window
 proc_agent_setup_window(Display ** dpy, int agent_number)
 {
   int i;
-  Window tmp_win;
+  Window tmp_win = 0;
 
-  /* Don't close but reopen...why this works? */
+  /* XCloseDisplay(*dpy); */
   *dpy = XOpenDisplay(NULL);	
   if (!*dpy) {
     fprintf(stderr, "Cannot open X Display(agent): %s\nABORTING!\n", 
             XDisplayName(NULL));
     return -1;
     }
+  XNoOp (*dpy);
 
   for (i = 0; i <= agent_number; i++)
     {
@@ -995,6 +1033,7 @@ dpscap_ConnSetup_display(xCAPConnSetupReq *req)
   unsigned char * msg = (unsigned char * )req;
   int off;
   char displayname[256];
+  gchar * env_pair;
   /* The display name might not be there and it certainly won't be NULL
      terminated.  We don't need to worry about overwriting what is after it */
   if (req->displayStringLength) 
@@ -1006,7 +1045,9 @@ dpscap_ConnSetup_display(xCAPConnSetupReq *req)
       displayname[req->displayStringLength] = '\0';
       if (DGS_DEBUG) 
         dprintf1("setting DISPLAY to '%s'\n", displayname);
-      setenv("DISPLAY", displayname, 1);
+      env_pair = g_strdup_printf("DISPLAY=%s", displayname);
+      putenv(env_pair);
+      g_free(env_pair);
     }
 }
 
@@ -1169,6 +1210,7 @@ dpscap_Notify_sync(gsdpsx_conn_info *conn, xCAPNotifyReq *req)
 
   event.xclient.display = xdev->dpy;
   event.xclient.window = conn->clientWindow;
+  if (0) g_message("[%s]", "dpscap_Notify_sync");
   XSendEvent(event.xclient.display, event.xclient.window, False, 0, &event);
   XSync(event.xclient.display, False);
 }
@@ -1234,9 +1276,9 @@ dpscap_Notify_pause_check(gsdpsx_conn_info *conn, int cxid, int sync)
       else
 	{
 	  resutl = XCheckIfEvent(shared_dpy, 
-			       &event,
-			       dpscap_Notify_pause_check_callback_x,
-			       (XPointer)cxid);
+				 &event,
+				 dpscap_Notify_pause_check_callback_x,
+				 (XPointer)cxid);
 	  if (resutl)
 	    conn->pause_cxids_set = g_set_remove(conn->pause_cxids_set,
 						 (gpointer)cxid);
@@ -1503,6 +1545,12 @@ gsdpsx_std_read_process(stream_state *st, stream_cursor_read *ignore_pr,
 						    (void *)cpsid);
   int newest_cpsid = gsdpsx_get_newest_context_index();
 
+      /* Context switing occured */
+      /* why context switching happens?  */
+      extern int reschedule_interval;
+      int ticks_left_value(void);
+
+
   if (pw->limit - pw->ptr < fork_op_length) 
     {
       /* Buffer is too small */
@@ -1510,6 +1558,7 @@ gsdpsx_std_read_process(stream_state *st, stream_cursor_read *ignore_pr,
 	dprintf("Stream buffer is too small.\n");
       return 1;
     }
+
   if (cpsid == conn->context_state_index_primary
       && conn->context_state_index_last != conn->context_state_index_primary)
     {
@@ -1525,12 +1574,10 @@ gsdpsx_std_read_process(stream_state *st, stream_cursor_read *ignore_pr,
     }
   else if (cpsid != conn->context_state_index_last)
     {
-      /* Context switing occured */
-      /* why context switching happens?  */
-      /* Reason: fork */
       dpscap_Notify_pause_check(&conn_info, 0, 0);
       if (conn->ctx_init_parameter.cxid != 0)
 	{
+	  /* Reason: fork */
 	  /* Associate gs_interp_context_state_current<->cxid */
 	  g_hash_table_insert(conn->cxid_to_cpsid_table,
 			      (void *)(conn->ctx_init_parameter.cxid),
@@ -1555,6 +1602,15 @@ gsdpsx_std_read_process(stream_state *st, stream_cursor_read *ignore_pr,
 	  gsdpsx_request_context_index(0);
 	  conn->context_state_index_last = cpsid;
 	}
+      else if (ticks_left_value() == (reschedule_interval - 1))
+	{
+	  if (0)
+	    g_message ("+ ticks: %d cpsid: %d last: %d", 
+		       ticks_left_value (), cpsid, conn->context_state_index_last);
+	  gsdpsx_request_context_index(conn->context_state_index_last);
+	  sched_act = SCHED_ACT_YIELD;
+	  goto forward;
+	}
     }
   else if (newest_cpsid && (conn->ctx_init_parameter.cxid != 0))
     {
@@ -1580,11 +1636,22 @@ gsdpsx_std_read_process(stream_state *st, stream_cursor_read *ignore_pr,
 					   (gpointer)cpsid);
       goto forward;
     }
+  else
+    {
+      int ticks_left_value(void);
+      if (0)
+	g_message ("- ticks: %d cpsid: %d last: %d", 
+		   ticks_left_value (), cpsid, conn->context_state_index_last);
+    }
+  if (0)
+    g_message ("cpsid: %d", cpsid);
+  conn->context_state_index_last = cpsid;
+  
   /* flush out any drawing commands in the queue */
   xdev = (gx_device_X *)gs_currentdevice(igs);
   if (!xdev)
     fprintf(stderr, "Cannot find device in gsdpsx_std_read_process\n");
-  if (xdev->dpy)
+  if (gx_device_is_x11(xdev) && xdev->dpy)
     XFlush(xdev->dpy);
 
   while (sched_act == SCHED_ACT_NOTHING
@@ -1665,7 +1732,7 @@ gsdpsx_std_read_process(stream_state *st, stream_cursor_read *ignore_pr,
 	buf_reset(buf);
 
       /* flush out any drawing commands in the queue */
-      if (xdev->dpy)
+      if (gx_device_is_x11(xdev) && xdev->dpy)
 	XFlush(xdev->dpy);
     } /* while data */
 
@@ -1695,7 +1762,7 @@ read_process_handle(gsdpsx_conn_info *conn,
 		    int remaining,
 		    enum scheduling_action * sched_act)
 {
-  int used, dealt_with = 1;
+  int used = 0, dealt_with = 1;
 
   if (pmsg[0] == DPSXOPCODEBASE) 
     {
@@ -1891,49 +1958,50 @@ gsdpsx_std_write_process(stream_state *st, stream_cursor_read *pr,
   gsdpsx_conn_info * conn = (gsdpsx_conn_info *)(s->file);
   static Atom outputAtom = 0;
   gx_device_X * xdev = (gx_device_X *)gs_currentdevice(igs);
-
-  if (!xdev || !xdev->dpy)
+  XEvent event;
+  int cpsid = gsdpsx_get_current_context_index();
+  CONTEXTXID cxid = 
+    (CONTEXTXID)g_hash_table_lookup(conn->cpsid_to_cxid_table,
+				    (void *)cpsid);
+  DPSCAPOutputEvent *out;
+  
+  if (!xdev || !gx_device_is_x11(xdev) || !xdev->dpy)
     return -1;
 
   if (!outputAtom)
     outputAtom = XInternAtom(xdev->dpy, DPSCAP_TYPE_PSOUTPUT_LEN, False);
+  
+  event.xclient.type = ClientMessage;
+  event.xclient.serial = 0;
+  event.xclient.send_event = True;
+  event.xclient.message_type = outputAtom;
+  event.xclient.format = 8;
+  event.xclient.display = xdev->dpy;
+  event.xclient.window = conn->clientWindow;
+  out = (DPSCAPOutputEvent*)event.xclient.data.l;
+
+  if (util_byteorder_check(conn))
+	  out->cxid = htonl(cxid);
+	else
+	  out->cxid = cxid;
 
   while ( count != 0 )
     {
       int towrite = min(count, DPSCAP_DATA_LEN);
-      XEvent event;
-      DPSCAPOutputEvent *out;
-
-      event.xclient.type = ClientMessage;
-      event.xclient.serial = 0;
-      event.xclient.send_event = True;
-      event.xclient.message_type = outputAtom;
-      event.xclient.format = 8;
-      out = (DPSCAPOutputEvent*)event.xclient.data.l;
-      {
-	int cpsid = gsdpsx_get_current_context_index();
-	CONTEXTXID cxid = (CONTEXTXID)g_hash_table_lookup(conn->cpsid_to_cxid_table,
-							  (void *)cpsid);
-	if (util_byteorder_check(conn))
-	  out->cxid = htonl(cxid);
-	else
-	  out->cxid = cxid;
-      }
       memcpy(out->data, pr->ptr+1, towrite);
       out->data[DPSCAP_DATA_LEN] = towrite;
-      event.xclient.display = xdev->dpy;
-      event.xclient.window = conn->clientWindow;
       if (0) {
         char data[21];
         memcpy(data,pr->ptr+1,towrite);
         data[towrite] = '\0';
         dprintf2("sent %d bytes of results\n  ->%s<-\n", towrite, data);
       }
-
+      if (0) g_message("[%s]", "gsdpsx_std_write_process");
       XSendEvent(event.xclient.display, event.xclient.window, False, 0, &event);
       pr->ptr += towrite;
       count -= towrite;
     }
+
   XSync(xdev->dpy, False);
   return 0;
 }
@@ -2648,7 +2716,9 @@ conn_server_create(void * sockaddr, enum conn_type type)
   if (type == UNIXCONN)
     {
       struct stat st;
-      (void)mkdir(CSDPS_UNIX_DIR, (S_IRWXU | S_IRWXG | S_IRWXO | S_ISVTX));
+      umask(0);
+      mkdir(CSDPS_UNIX_DIR, (S_IRWXU | S_IRWXG | S_IRWXO));
+      chmod(CSDPS_UNIX_DIR, (S_IRWXU | S_IRWXG | S_IRWXO | S_ISVTX));
 
       if (stat(CSDPS_UNIX_DIR, &st)) {
         perror("stat: ");
@@ -2954,7 +3024,7 @@ conn_atom_cleanup_window(Window win, Window * data, int numAgents)
 private int
 conn_insock_accept(int server_socket, enum conn_type type)
 {
-  void * in_sockaddr;
+  void * in_sockaddr = NULL;
   struct sockaddr_un in_sockaddr_un;
   struct sockaddr_in in_sockaddr_in;
   int in_sock, addrlen;
@@ -3285,7 +3355,7 @@ util_signal_int2str(int sig)
 }
 
 private void
-util_print_pid(int sec)
+debug_print_pid(int sec)
 {
   int i;
 
@@ -3357,7 +3427,7 @@ status_send(gsdpsx_conn_info *conn, int cpsid, int status)
     (CONTEXTXID)g_hash_table_lookup(conn->cpsid_to_cxid_table,
 				    (void *)cpsid);
   
-  if (!xdev || !xdev->dpy)
+  if (!xdev || !gx_device_is_x11(xdev) || !xdev->dpy)
     return;
 
   if (!statusAtom)
@@ -3383,6 +3453,7 @@ status_send(gsdpsx_conn_info *conn, int cpsid, int status)
     }
   event.xclient.display = xdev->dpy;
   event.xclient.window = conn->clientWindow;
+  if (0) g_message("[%s]", "status_send");
   XSendEvent(event.xclient.display, event.xclient.window, False, 0, &event);
   XSync(xdev->dpy, False);
 }
